from math import ceil
from collections import deque
import tkinter as tk

class Diagram(tk.Frame):
    "A diagram that changes dynamically."

    def __init__(self, parent, maxlen=None, fg=None, thick=1, **kw):
        "Initialise object; maxlen is max number of values of coordinates; if none then unbounded."
        super().__init__(parent)
        self.ca = tk.Canvas(self)
        self.ca.grid(sticky="wesn")
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.config(**kw)
        self.xx = deque(maxlen=maxlen)
        self.yy = deque(maxlen=maxlen)
        self.iline = None     #The previous line (initialy None, as no line has been drawn)
        self.ca.bind("<Configure>", self.regen)    # Bind in the end, in order to avoid preemptive calls
                                                   # It calls regen with an event

    def config(self, **kw):
        "Pass all the configuration options to the canvas object."
        self.fill = kw.pop("fg", None)
        self.thick = kw.pop("thick", None)
        self.ca.config(**kw)


    def destroy(self):
        "Remove circular references."
        self.ca.destroy()
        del self.ca
        super().destroy()


    def add(self, x, y):
        "Add a new point and change the view."
        self.xx.append(x)
        self.yy.append(y)
        self.regen(None)       #None is the event (which is not used)


    def regen(self, event):
        "Rebuild the diagram."
        if len(self.xx) < 2: return     #only 1 point: can not draw line
        xmin = min(self.xx)
        xmax = max(self.xx)
        ymin = min(self.yy)
        ymax = max(self.yy)
        dx = xmax-xmin
        dy = ymax-ymin
        if dx <= 0.0 or dy <= 0.0: return  #can not calculate scale
        self.ca.update()                      # So that the next 2 instructions will probably succeed
        w = self.ca.winfo_width()             # Pixels
        h = self.ca.winfo_height()            # Pixels
        t = 1 if self.thick is None else self.thick
        margin1 = 1 + int(ceil(t/2))    # upper and lower margin in pixels
        if w <= 2*margin1 or h <= 2*margin1: return  #either the widget is too small, or tkinter did not reliably report the width and height
        pklx = (w-1)/dx
        pkly = (h-1-2*margin1)/dy
        px = [int((x-xmin)*pklx) for x in self.xx]
        py = [h-1-(margin1+int((y-ymin)*pkly)) for y in self.yy]
        icod = self.ca.create_line(list(zip(px, py)), fill=self.fill, width=self.thick)
        self.ca.delete(self.iline)
        self.iline = icod
        self.ca.update()


def test():
    "Test Diagram."
    from math import sin
    import random

    def addp():
        "Add a point to the diagram."
        dia.add(point[0], point[1])
        point[0] += 0.1
        point[1] = sin(point[0])
        if random.random() > 0.98: point[1] = random.random()*2*sin(point[0])
        dia.after(100, addp)

    point = [0.0, sin(0.0)]
    root = tk.Tk()
    dia = Diagram(root, maxlen=200, fg="yellow", thick=10, bg="green")
    dia.grid(sticky="wesn")
    but = tk.Button(root, text="Add point", command=addp)
    but.grid()
    root.columnconfigure(0, weight=1)
    root.rowconfigure(0, weight=1)
    root.mainloop()


if __name__ == "__main__":
    test()
