module modanneal
use modparam
!use modtyp
use modmath
use modroomconfiguration
implicit none
private
public :: SimulatedAnnealing, init, anneal


type SimulatedAnnealing
!     "The class implements the simulated annealing method."
      integer(KIL) :: nTsteps   ! = 30;   max temperature steps
      integer(KIL) :: nChan     ! = 100;  max number of state changes tried at any temperature
      integer(KIL) :: nSch      ! = 10;   max number of succesful state changes

      integer(KIL) :: iSch      ! = 0;    number of succesful state changes (positive or negative)
      integer(KIL) :: iSchNeg   ! = 0;    number of succesful energy-negative state changes

      integer(KIL) :: nDimState ! = 20;   The dimension of the vector space which each state
                                !         belongs to. Note that it may be not constant.

      real(KRL) :: t       ! = 100.0; temperature
      real(KRL) :: tFactr  ! = 0.70;  Annealing schedule: t is reduced by this factor on each step
      real(KRL) :: e1      ! = 0.0;   energy before and ..
      real(KRL) :: e2      ! = 0.0;   after a change
      real(KRL) :: emin    ! = 0.0;   minimum energy found
      type(Random) :: r    ! = Random();
end type SimulatedAnnealing

interface init
    module procedure SimulatedAnnealingInit
end interface init

interface anneal
    module procedure SimulatedAnnealingAnneal
end interface anneal

contains

  
subroutine SimulatedAnnealingInit(self)
!     """Initialize simulated annealing procedure.
!      
!     These values are typical. They will be overwritten by anneal()."""
      type(SimulatedAnnealing), intent(inout) ::  self
      self%nTsteps  = 30   ! max temperature steps
      self%nChan    = 100  ! max number of state changes tried at any temperature
      self%nSch     = 10   ! max number of succesful state changes
 
      self%iSch     = 0    ! number of succesful state changes (positive or negative)
      self%iSchNeg  = 0    ! number of succesful energy-negative state changes

      self%nDimState = 20  ! The dimension of the vector space which each state
                           !  belongs to. Note that it may be not constant.

      self%t      = 100.0_KRL ! temperature
      self%tFactr = 0.70_KRL  ! Annealing schedule: t is reduced by this factor on each step
      self%e1     = ZERO      ! energy before and ..
      self%e2     = ZERO      ! after a change
      self%emin   = ZERO      ! minimum energy found
      call init(self%r)
end subroutine SimulatedAnnealingInit

subroutine SimulatedAnnealingAnneal(self, obj)
!     "Does the annealing process."
!     Initial average Delta Energy (just run the algorithm for ndimstate times, calculate 
!     delta energy and take the average) should be normalised to about 100
      type(SimulatedAnnealing), intent(inout) ::  self
      type(RoomConfiguration), intent(inout) ::  obj
      integer(KIL) :: iTstep
      real(KRL) :: eprev
      character(len=79) :: dl
      self%nTsteps = 50        ! max temperature steps
      self%tFactr  = 0.70_KRL  ! Annealing schedule: t is reduced by this factor on each step
      self%t = 100.0_KRL       ! initial temperature

!-----Initial sharing

      self%nDimState = initState(obj)
      self%e1 = energyState(obj)
      self%emin = self%e1
      eprev = self%e1
      call saveMinState(obj)
      call prg(" ")
      call prg("")
      write (dl, 10)"", " ", "", "", " .", "max ."
10    format (a3, 2x, a8, 2x, a6, 2x, a6, 2x, a6, a8, 2x, a8)
      call prg(dl)

!-----Temperature loop-------------------------------------------

      do iTstep=1, self%nTsteps
          call annealTrials(self, obj)
          write (dl, 20) iTstep, self%t, self%iSch, self%iSchNeg, self%iSch-self%iSchNeg, &
                         eprev-self%e1, eprev-self%emin
20        format (i3, f10.5, i8, i8, i8, 2f10.3)
          call prg(dl)
          if (self%iSch == 0) exit     ! If no succesful changes, end.
          self%t = self%t * self%tFactr
          eprev = self%e1
      end do
      self%nDimState = restoreMinState(obj)
end subroutine SimulatedAnnealingAnneal


subroutine AnnealTrials(self, obj)
!     "Annealling fro 1 temperature."
      type(SimulatedAnnealing), intent(inout) ::  self
      type(RoomConfiguration), intent(inout) ::  obj
      integer(KIL) :: nDsMax
      integer(KIL) :: iChan

!-----Trial sharings' loop

      self%nChan  = 200 * self%nDimState  ! max number of state changes tried at any temperature
      self%nSch   =  20 * self%nDimState  ! max number of succesful bananas changes
      nDsMax = self%nDimState             ! maximum dimension at this temperature

      self%iSch    = 0     ! number of succesful state changes (positive or negative)
      self%iSchNeg = 0     ! number of succesful energy-negative state changes
      iChan = 0

      do while (.TRUE.)
          iChan = iChan + 1
!         print *, "iChan=", iChan
          if (iChan > self%nChan) return

          self%nDimState = changeState(obj)
          self%e2 = energyState(obj)
!         print *, "e1,e2=, self%e1, self%e2
          if (metrop(self, self%e2-self%e1, self%t)) then
              if (self%e2 < self%emin) then
                  self%emin = self%e2
                  call saveMinState(obj)
              end if
              if (self%nDimState > nDsMax) then
                  self%nChan = 100 * self%nDimState
                  self%nSch  =  10 * self%nDimState
                  nDsMax = self%nDimState
              end if

              if (self%e2-self%e1 <  ZERO) self%iSchNeg = self%iSchNeg + 1
              if (self%e2-self%e1 /= ZERO) self%iSch = self%iSch + 1
              self%e1 = self%e2
              if (self%iSch > self%nSch) return
          else
              self%nDimState = unChangeState(obj)
          end if
      end do
end subroutine AnnealTrials


logical function metrop(self, de, t)
!     "Metropolis algorithm."
      type(SimulatedAnnealing), intent(inout) ::  self
      real(KRL), intent(in) :: de, t
      real(KRL) :: x, xm
!     print *, "metrop de, de/t=", de, de/t
      if (de < ZERO) then
          metrop = .TRUE.
          return
      end if
      x = real1(self%r)
      xm = exp(-de/t)
!     print *, "metrop xm,x=", xm, x
      metrop = (x < xm)
end function metrop

end module modanneal
