module modroomconfiguration
!This module describes an annealable object
use modparam
use modmath
use modtyp
use modroomtype
use modroom
use modthanroomlist

implicit none
private
public :: RoomConfiguration, init, free, plot, initState, energyState, &
          saveMinState, restoreMinState, &
          saveOldState, changeState, unChangeState

type RoomConfiguration
        type(Room), pointer :: root, rootOld, rootMin
	real(KRL) :: per, efact
	type(Random) :: r
	type(ThanRoomList) :: all, mable, sable, chable
end type RoomConfiguration

interface init
    module procedure roomConfigurationInit
end interface init

interface free
    module procedure roomConfigurationFree
end interface free

interface plot
    module procedure RoomConfigurationPlot
end interface plot

interface initState
    module procedure roomconfigurationinitState
end interface initState

interface energyState
    module procedure roomconfigurationenergyState
end interface energyState

interface saveMinState
    module procedure roomconfigurationsaveMinState
end interface saveMinState

interface restoreMinState
    module procedure roomconfigurationrestoreMinState
end interface restoreMinState

interface saveOldState
    module procedure roomconfigurationsaveOldState
end interface saveOldState

interface changeState
    module procedure roomconfigurationchangeState
end interface changeState

interface unChangeState
    module procedure roomconfigurationunChangeState
end interface unChangeState

contains

subroutine roomConfigurationInit(self, b, h)
!       "Initial configuration is a single room."
        type(RoomConfiguration), intent(inout) :: self
        real(KRL), intent(in) :: b, h
        real(KRL), dimension(4) :: xymm
        xymm(1) = ZERO
        xymm(2) = ZERO
        xymm(3) = b
        xymm(4) = h
        allocate (self%root)
        call roomInit(self%root, xymm)
	nullify (self%rootOld, self%rootMin)
	self%per = 0.2                  ! Do not split nearer to 20% from sides
	call randInit(self%r)
	call init(self%all)
	call init(self%mable)
	call init(self%sable)
	call init(self%chable)
	call typeNodes(self) 
end subroutine roomConfigurationInit


subroutine roomConfigurationFree(self)
!       "Delete configuration"
        type(RoomConfiguration), intent(inout) :: self
	call free(self%all)
	call free(self%mable)
	call free(self%sable)
	call free(self%chable)
	call roomFree(self%root)
	deallocate(self%root)
	nullify(self%root)
	if (associated(self%rootOld)) then
	    call roomFree(self%rootOld)
            deallocate(self%rootOld)
            nullify(self%rootOld)
        end if
	if (associated(self%rootMin)) then
	    call roomFree(self%rootMin)
            deallocate(self%rootMin)
            nullify(self%rootMin)
        end if
end subroutine roomConfigurationFree


real(KRL) function RoomConfigurationEnergyState(self)
!       "Computes the energy of the configuration."
        type(RoomConfiguration), intent(inout) :: self
        integer(KIL) :: i
        real(KRL) :: e
	type(Room), pointer :: room1
	e = ZERO
	do i=1, llen(self%sable)
	     room1 => get(self%sable, i)
	     e = e + roomEnergy(room1)
	end do
	if (llen(self%sable) > 5) e = e + (llen(self%sable)-5)*200
	RoomConfigurationEnergyState = self%efact * e / llen(self%sable)
end function RoomConfigurationEnergyState

integer(KIL) function RoomConfigurationChangeState(self)
!       "Changes the configuration a little."
        type(RoomConfiguration), intent(inout) :: self
        integer(KIL) :: i
        real(KRL) :: p, dc, px, dx, dy, c
        type(Room), pointer :: room1
        call saveOldState(self)
	p = randReal1(self%r)
	if (p < 0.4 .and. llen(self%chable)>0) then    ! In case that only the root room exists
	    i = randint(self%r, 1, llen(self%chable))
!	    print *, "change", llen(self%chable), i
	    room1 => get(self%chable, i)
	    i = room1%isplit
	    dc = (room1%xymm(i+2) - room1%xymm(i))*self%per
	    room1%csplit = randUniform(self%r, room1%xymm(i)+dc, room1%xymm(i+2)-dc)
	    room1%roomll%xymm(i+2) = room1%csplit
	    room1%roomur%xymm(i)   = room1%csplit
	else if (p < 0.7 .or. llen(self%chable) == 0) then ! .and. llen(self%sable) <=5)   ! If none is mergeable then split
	    i = randint(self%r, 1, llen(self%sable))
!            print *, "split", llen(self%sable), i
	    room1 => get(self%sable, i)
!	    print *, "room1=", room1%xymm
	    dx = room1%xymm(3)-room1%xymm(1)
	    dy = room1%xymm(4)-room1%xymm(2)
	    px = 1.0/(1.0+dy/dx)
!	    print *, "dx, dy, px", dx, dy, px
	    if (randReal1(self%r) < px) then
	        i = 1
	    else
	        i = 2
	    end if
	    dc = (room1%xymm(i+2) - room1%xymm(i))*self%per
	    c = randUniform(self%r, room1%xymm(i)+dc, room1%xymm(i+2)-dc)
	    call roomSplit(room1, i, c)
	    call typeNodes(self)
	else
            !print *, "merge"
	    i = randint(self%r, 1, llen(self%chable))
	    room1 => get(self%chable, i)
	    call roomMerge(room1)
	    call typeNodes(self)
	end if
	RoomConfigurationChangeState = llen(self%all)       ! Dimensionality of the configuration
end function RoomConfigurationChangeState

subroutine RoomConfigurationSaveOldState(self)
!       "Save current configuration for future unChangeState()."
        type(RoomConfiguration), intent(inout) :: self
        if (associated(self%rootOld)) call roomFree(self%rootOld)
        self%rootOld => roomClone(self%root)
end subroutine RoomConfigurationSaveOldState

integer(KIL) function RoomConfigurationUnChangeState(self)
!       "Restores previously saved state."
        type(RoomConfiguration), intent(inout) :: self
        !print *, "unchanged"
        call roomFree(self%root)
	self%root => self%rootOld
	nullify (self%rootOld)
	call typeNodes(self)
	RoomConfigurationUnChangeState = llen(self%all)       ! Dimensionality of the configuration
end function RoomConfigurationUnChangeState


subroutine RoomConfigurationSaveMinState(self)
!       "Save current configuration as the best so far."
        type(RoomConfiguration), intent(inout) :: self
        if (associated(self%rootMin)) call roomFree(self%rootMin)
        self%rootMin => roomClone(self%root)
end subroutine RoomConfigurationSaveMinState


integer(KIL) function RoomConfigurationRestoreMinState(self)
!       "Restores previously saved best configuration."
        type(RoomConfiguration), intent(inout) :: self
        call roomFree(self%root)
	self%root => self%rootMin
	nullify (self%rootMin)
	call typeNodes(self)
	RoomConfigurationRestoreMinState = llen(self%all)       ! Dimensionality of the configuration
end function RoomConfigurationRestoreMinState


integer(KIL) function RoomConfigurationInitState(self)
!     "Initialize random changes and calibrate energy."
      type(RoomConfiguration), intent(inout) :: self
      integer(KIL) :: ndim, tries, i, j
      real(KRL) :: e1, de, e2
      character(len=120) :: dl
      self%efact = 1.0
      ndim = 5                        ! Assume 5 rooms in house	  
      tries = ndim*10
      do i=1, tries                   ! Randomise a bit
          print *, i
          j = RoomConfigurationchangeState(self)
      end do
      print *, "len=", llen(self%all)
      e1 = RoomConfigurationEnergyState(self)
      de = ZERO
      ndim = llen(self%sable)
      tries = ndim*10 
      do i=1, tries
          !print *, i
          j = RoomConfigurationchangeState(self)
          e2 = RoomConfigurationenergyState(self)
          de = de + abs(e2-e1)
          e1 = e2
      end do
      de = de / tries
      self%efact = 100.0 / de             ! Normalise delta energy to 100: efact*de = 100
      write (dl, 10) e1
10    format (" =", f15.3)
      call prg(dl)
      write (dl, 20)
20    format ("   =", f15.3)
      call prg(dl)
      print *, "number of rooms:", llen(self%all)
      print *, "efact=", self%efact
!     call RoomConfigurationplot(self, "First approximation")
      RoomConfigurationInitState = llen(self%all)
end function RoomConfigurationInitState

subroutine typeNodes(self)
!     "Finds a list of splittable, a list of mergeable nodes and a list of all."
      type(RoomConfiguration), intent(inout) :: self
      call clear (self%all)
      call clear (self%mable)
      call clear (self%sable)
      call clear (self%chable)
      call roomList(self%root, self%all, self%mable, self%sable, self%chable)
end subroutine typeNodes

subroutine RoomConfigurationPlot(self, tit)
!     Plots the room configuration (just write the lines to a file)."
      type(RoomConfiguration), intent(inout) :: self
      character(len=*), intent(in) :: tit
      integer(KIL), parameter :: ch=20
      open (ch, file="plot.txt")
      write (ch, 10) tit
10    format (a)
      call roomPlot(self%root, ch, 0)
      close (ch)
end subroutine RoomConfigurationPlot

end module modroomconfiguration
