module modroom
use modparam
use modmath
use modroomtype
use modthanroomlist
implicit none
public
private :: penalty_width

type RoomConstrain
    real(KRL) :: bmin1, bmin, bmax, bmax1
    real(KRL) :: penalty1, penalty
    integer(KIL) :: maxrooms
    real(KRL) :: penaltymore
end type RoomConstrain


contains

subroutine roomInit(self, xymm)
!       "The xymm are the coordinates of lower left followed by the coordinates of upper right."
        type(Room), intent(inout) :: self
        real(KRL), dimension(4), intent(in) :: xymm 
        integer(KIL) :: i
        self%xymm = xymm
	self%isplit = -1                   ! Room is not (yet) split
        self%csplit = -1.0e30_KRL          ! Invalid value
	nullify (self%roomll, self%roomur) ! Children
end subroutine roomInit

recursive subroutine roomFree(self)
!       "The xymm are the coordinates of lower left followed by the coordinated of upper right."
        type(Room), intent(inout) :: self
        if (associated(self%roomll)) then
            call roomFree(self%roomll)
	    deallocate(self%roomll)
	    nullify (self%roomll)
	end if
        if (associated(self%roomur)) then
            call roomFree (self%roomur)
            deallocate (self%roomur)
            nullify (self%roomur)
        end if
end subroutine roomFree

recursive function roomClone(self) result (roomCloned)
!       "Makes a copy of itsef."
        type(Room), pointer :: roomCloned
        type(Room), intent(inout) :: self
	type(Room), pointer :: roomnew
	allocate (roomnew)
        call roomInit(roomnew, self%xymm)
	if (self%isplit == -1) then
	    roomCloned => roomnew
	    return
	end if
	roomnew%isplit = self%isplit
	roomnew%csplit = self%csplit
	if (associated(self%roomll)) roomnew%roomll => roomClone(self%roomll)
	if (associated(self%roomur)) roomnew%roomur => roomClone(self%roomur)
	roomCloned => roomnew
end function roomClone

subroutine roomSplit(self, isplit, csplit)
!       "Split the room to 2 children."
        type(Room), intent(inout) :: self
        integer(KIL), intent(in) :: isplit
        real(KRL) , intent(in) :: csplit
        real(KRL), dimension(4) :: xymm
        if (.not.roomIsleaf(self)) stop "room is already split!"
	if (.not.(self%xymm(isplit) < csplit .and. csplit < self%xymm(isplit+2))) &
	    stop "Split line outside of the room"
	self%isplit = isplit
	self%csplit = csplit
        xymm = self%xymm
	xymm(isplit+2) = csplit
	allocate (self%roomll)
	call roomInit(self%roomll, xymm) 
        xymm = self%xymm
	xymm(isplit) = csplit
	allocate (self%roomur)
	call roomInit(self%roomur, xymm) 
end subroutine roomSplit

subroutine roomMerge(self)
!       "Merges all the children/grandchildren of the room."
        type(Room), intent(inout) :: self
        if (associated(self%roomll)) call roomFree(self%roomll)
        if (associated(self%roomur)) call roomFree(self%roomur)
	self%isplit = -1
	self%csplit = -1.0e30_KRL
	nullify (self%roomll, self%roomur)
end subroutine roomMerge

logical function roomIsleaf(self)
!       "Checks if the room is a leaf node (it has not been split)."
        type(Room), intent(inout) :: self
        roomIsleaf = self%isplit == -1
end function roomIsleaf

recursive subroutine roomList(self, all, mall, sall, chall)
!       "Returns a list of self and children."
        type(Room), intent(inout) :: self
        type(ThanRoomList), intent(inout) :: all, mall, sall, chall
	call append(all, self)
	if (roomIsleaf(self)) then
	    call append(sall, self)
	else
	    if (self%roomll%isplit == -1 .and. &
	        self%roomur%isplit == -1) call append(chall, self)
	    call append(mall, self)
	    call roomList(self%roomll, all, mall, sall, chall)
	    call roomList(self%roomur, all, mall, sall, chall)
	end if
end subroutine roomList


real(KRL) function roomEnergy(self, rcon)
!       "Computes the energy of the room."
        type(Room), intent(inout) :: self
        type(RoomConstrain), pointer :: rcon
        real(KRL) :: e
	e  =     penalty_width(self%xymm(3)-self%xymm(1), rcon)
	e  = e + penalty_width(self%xymm(4)-self%xymm(2), rcon)
	roomEnergy = e
end function roomEnergy


subroutine roomPlot(self, ch, icol)
!       "Plots the room into a ThanChart."
        type(Room), intent(inout) :: self
        integer(KIL), intent(in) :: ch, icol
        real(KRL) :: xa, ya, xb, yb
        write (ch, 10) icol
10      format ("LINE_COLOR", i4)
        xa = self%xymm(1)
	ya = self%xymm(2)
	xb = self%xymm(3)
	yb = self%xymm(4)
        write (ch, 20) xa, ya
20      format (2e15.8) 
        write (ch, 20) xb, ya
        write (ch, 20) xb, yb
        write (ch, 20) xa, yb
        write (ch, 20) xa, ya
        if (.not.roomIsleaf(self)) call roomPlotSplit(self, ch, icol+1)
end subroutine roomPlot

recursive subroutine roomPlotSplit(self, ch, icol)
!       "Plots the split line of the room."
        type(Room), intent(inout) :: self
        integer(KIL), intent(in) :: ch, icol
        integer(KIL) :: ic
        real(KRL) :: xa, ya, xb, yb
        if (self%isplit == 1) then
	    xa = self%csplit
	    xb = self%csplit
	    ya = self%xymm(2)
	    yb = self%xymm(4)
	else
	    xa = self%xymm(1)
	    xb = self%xymm(3)
	    ya = self%csplit
	    yb = self%csplit
	end if
        write (ch, 10) icol
10      format ("LINE_COLOR", i4)
        write (ch, 20) xa, ya
20      format (2e15.8) 
        write (ch, 20) xb, yb
        ic = icol
	if (.not.roomIsleaf(self%roomll)) then
	    call roomPlotSplit(self%roomll, ch, ic+1)
	    ic = ic + 1
	end if
	if (.not.roomIsleaf(self%roomur)) call roomPlotSplit(self%roomur, ch, ic+1)
end subroutine roomPlotSplit

real(KRL) function penalty_width(b, c)
!       "Computes the penalty due to small or big width."
        real(KRL), intent(in) :: b
        type(RoomConstrain), pointer :: c
	real(KRL) :: e
        if (b <= ZERO) stop "b<=0"
        if (b < c%bmin1) then
	    e = (c%bmin-c%bmin1)*c%penalty + (c%bmin1-b)*c%penalty1
	else if (b < c%bmin) then
	    e = (c%bmin-b)*c%penalty
	else if (b <= c%bmax) then
	    e = ZERO
	else if (b < c%bmax1) then
	    e = c%bmax1-b
	else
	    e = (c%bmax1-c%bmax)*c%penalty + (b-c%bmax1)*c%penalty1
	end if
	penalty_width = e
end function penalty_width

real(KRL) function penalty_width_old(b)
!       "Computes the penalty due to small or big width."
        real(KRL), intent(in) :: b
	real(KRL) :: e
        if (b <= ZERO) stop "b<=0"
        if (b < 2.5) then
	    e = 3-2.5 + (2.5-b)*10
	elseif (b < 3.0_KRL) then
	    e = 3-b
	elseif (b <= 4) then
	    e = ZERO
	else if (b < 5) then
	    e = 5-b
	else
	    e = 5-4 + (b-5)*10
	end if
	penalty_width_old = e
end function penalty_width_old


end module modroom
