# -*- coding: iso-8859-7 -*-
##############################################################################
# ThanCad 0.2.2 "Urban SAR": 2dimensional CAD with raster support for engineers.
# 
# Copyright (c) 2001-2013 Thanasis Stamos,  January 16, 2013
# URL:     http://thancad.sourceforge.net
# e-mail:  cyberthanasis@excite.com
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (www.gnu.org/licenses/gpl.html).
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##############################################################################

"""\
ThanCad 0.2.2 "Urban SAR": 2dimensional CAD with raster support for engineers.

Package which creates a bioclimatic city plan.
This module builds a cache for an Hippodamian plan of city layouts.
"""

from math import cos, sin, ceil
from p_gmath import dpt
import p_ggen, p_gtri
import roadut, geomcirc


class CircularCache(object):
    "A cache of road energys for a circular urban plan system."

    def __init__(self, dr=1.0, dxy=10.0, factex=1.0, devdebug=False):
        "Initialize the cache."
        self.dr = dr                                               #Sampling Distance x and y of all computed roads in the cache
        self.dxy = dxy                                             #Sampling x and y for the centers
        self.factex = factex                                       #Multiple of hull size that the hull will be extened in all directions
        self.dtheta = None                                         #Sampling theta for computation of road cache
        self.xymm = None                                           #Enclosing reactangle, where the centers are confined
        self.energycache = None                                    #The cache: energy of rings and rays for every center
        self.devdebug = devdebug                                   #If True, execute Developer debug code
        self.hull = None                                           #The city plan region


    def shallowClone(self):
        "Make a shallow copy of self."
        raise IndexError
        r = HippoCache(self.dx, self.dy, self.dtheta)
        r.roadenx  = self.roadenx
        r.roadeny  = self.roadeny
        r.pxmin    = self.pxmin
        r.pxmax    = self.pxmax
        r.pymin    = self.pymin
        r.pymax    = self.pymax
        return r


    def build_cache(self, dtm, hull, fw=None, prt=p_ggen.prg):
        "Create road grid every dx, dy, dtheta; cache the results; boik1 is the minimum city block width in the circumferential direction."
        self.hull = [list(cc) for cc in hull]
        xs = [cc[0] for cc in hull]
        ys = [cc[1] for cc in hull]
        xymm = [min(xs), min(ys), max(xs), max(ys)]
        dx = xymm[2] - xymm[0]
        dy = xymm[3] - xymm[1]
        siz = max(dx, dy)
        siz = siz * self.factex
        self.xymm = [xymm[0]-siz, xymm[1]-siz, xymm[2]+siz, xymm[3]+siz]
        del xymm
        for i,c in enumerate(self.xymm):
            self.xymm[i] = round(c/self.dxy)*self.dxy
        self.dtheta = self.dr / siz

        if fw is not None:
            write_cache_gen(fw, self.dr, self.dxy, self.dtheta, self.xymm, self.hull)
        for cen in cens:
            rings, closedrings, rays, enring, enray = self.energycache[cen]
            write_cache_center1(fw, cen, rings, closedrings, rays, enring, enray)


        self.energycache = {}
        n = int((self.xymm[2]-self.xymm[0])/self.dxy) * int((self.xymm[3]-self.xymm[1])/self.dxy)
        i = 0
        prt("Building energy cache for center:")
        for xc in p_ggen.xfrange(self.xymm[0], self.xymm[2], self.dxy):
            for yc in p_ggen.xfrange(self.xymm[1], self.xymm[3], self.dxy):
                cen = xc, yc
                i += 1
                prt("Center %15.3f%15.3f  (%d/%d)" % (xc, yc, i, n))
                rays = geomcirc.radials(cen, hull, self.dtheta, self.dr)
                rings, closedrings = geomcirc.arcsx(rays, self.dtheta, self.dr)
                crings, crays = addz(dtm, cen, rings, rays)
                enring, enray = penalties(dtm, rings, rays, crings, crays)
                if fw is not None:
                    write_cache_center1(fw, cen, rings, closedrings, rays, enring, enray)

                self.energycache[cen] = rings, closedrings, rays, enring, enray
#                if self.devdebug:
#                    import p_gdxf
#                    dxf = p_gdxf.ThanDxfPlot()
#                    dxf.thanDxfPlots()
#                    geomcirc.plotcen(dxf, cen, self.hull, rays, rings)
#                    dxf.thanDxfPlot(0,0,999)
                break
            break


    def energy(self, s):
        "Calculates the energy of the road grid described in state s, using the built cache."
        rings, rays = self.energycache[s.cen]
        sd, sd8, sd10, se = 0.0
        for r in s.rs:
            for pline in rings[r]:
                d, d8, d10, e = rays[theta]
                sd += d
                sd8 += d8
                sd10 += d10
                se += e
        for theta in s.thetas:
            d, d8, d10, e = rays[theta]
            sd += d
            sd8 += d8
            sd10 += d10
            se += e
        return se/sd


    def energy3(self, s):
        "Calculates the energy of the road grid described in state s, using the built cache; returns more detail."
        sd, sd8, sd10, se = self.roadenergy(self.roadenx, self.dy, s.theta, self.pymin, self.pymax, s.boik, s.bod)
        d, d8, d10, e     = self.roadenergy(self.roadeny, self.dx, s.theta, self.pxmin, self.pxmax, s.hoik, s.bod)
        sd += d
        sd8 += d8
        sd10 += d10
        se += e
        return se/sd, sd8, sd10


    def roadenergy(self, s):
        "Find the energy of roads along x or y axis using the built cache."
        rings, closedrings, rays, enring, enray = self.energycache[s.cen]
        thetamin, _ = rays[0]
        thetamax, _ = rays[-1]
#        print "thetamin, max=", thetamin, thetamax, self.dtheta
        rmin, _ = rings[0]
        rmax, _ = rings[-1]
#        print "rmin, max=", rmin, rmax

        thpot = []                 #Potential thetas of the city plan (sparse)
        dis = 0.0
        j = jj = 0
        while True:
            dis += s.bod[jj]/2.0
            th = ceil((thetamin + dis/rmax)/self.dtheta)*self.dtheta
            th = round(th, 8)                             #Numerical reasons
            if th > thetamax: break
            thpot.append(th)
            dis += s.bod[jj]/2.0 + s.boik[jj]
            j += 1
            if j < len(s.boik): jj = j

        rpot = []                  #Potential r's of the cityplan (sparse)
        dis = rmin
        j = jj = 0
        while True:
            dis += s.bod[jj]/2.0
            if dis > rmax: break
            r = ceil(dis/self.dr)*self.dr
            rpot.append(r)
            dis += s.bod[jj]/2.0 + s.hoik[jj]
            j += 1
            if j < len(s.hoik): jj = j

        rpot.reverse()
        rpotth = {}                #Potential rs for each theta (even sparser)
        ths = list(thpot)
        for r in rpot:             #For each (diminishing) r, ensure that we have the minimum length..
            j = 1                  #.. b of city block. When a theta is deleted, it is deleted for..
            while j < len(ths):    #.. all subsequent (smaller) r's
                if (ths[j]-ths[j-1])*r >= s.bod1 + s.boik1:
                    j += 1
                else:
                    del ths[j]
#            print r, closedrings[r]
            if closedrings[r]:
#                print "delta theta:", ths[0]-ths[-1], dpt(ths[0]-ths[-1])
                if dpt(ths[0]-ths[-1])*r < s.bod1 + s.boik1:
                    del ths[-1]
            for th in ths:
                rpotth.setdefault(th, set()).add(r)

        actrays = []               #Actual rays of the city plan
        for theta, ray in rays:
            if theta not in rpotth: continue
            rset = rpotth[theta]
            actray = [r for r in ray if r in rset]
            if len(actray) > 1: actrays.append((theta, actray))

        sd = sd8 = sd10 = se = 0.0
        for th, ray in actrays:                       #Compute energy of rays
#            print th, ray[0], ray[-1]
#            print sorted(enray.keys())
#            print sorted(enray[th].keys())
            d,  d8,  d10,  e  = enray[th][ray[0]]
            db, d8b, d10b, eb = enray[th][ray[-1]]
            sd   += db   - d
            sd8  += d8b  - d8
            sd10 += d10b - d10
            se   += eb   - e

        for r in rpot:                                #Compute energy of rings. The potential..
            d, d8, d10, e = enring[r]                 #.. r's here, are all present
            sd   += d
            sd8  += d8
            sd10 += d10
            se   += e

        if self.devdebug:
            actrings = [(r, thetasegments) for r, thetasegments in rings if r in rpot]
            import p_gdxf
            dxf = p_gdxf.ThanDxfPlot()
            dxf.thanDxfPlots()
            geomcirc.plotcen(dxf, s.cen, self.hull, actrays, actrings)
            dxf.thanDxfPlot(0,0,999)
        return sd, sd8, sd10, se


    def roundt(self, theta):
        "Round to fit the cache."
#        if self.USEDEM: return theta
        dth = self.dtheta
        th = round(theta/dth)*dth
        return round(th, 6)


    def write_cache(self, fw):
        "Write the cache to a file (like) object."
        write_cache_gen(fw, self.dr, self.dxy, self.dtheta, self.xymm, self.hull)
        cens = sorted(self.energycache)
        for cen in cens:
            rings, closedrings, rays, enring, enray = self.energycache[cen]
            write_cache_center1(fw, cen, rings, closedrings, rays, enring, enray)


    def read_cache(self, it):    #May raise ValueError, IndexError
        "Read HippoUrban from a file (like) object."
        self.__init__()
        self.dr, self.dxy, self.dtheta = map(float, next(it).split())
        self.xymm = map(float, next(it).split())
        if len(self.xymm) != 4: raise ValueError("Exactly 4 coordinates were expected")
        self.hull = []
        while True:
            dline = next(it)
            if dline == "$": break
            cc = map(float, dline.split())
            if len(cc) < 2: raise ValueError("2 or more coordinates were expected")
            self.hull.append(cc)

        self.energycache = {}
        for dline in it:
            cen = tuple(map(float, dline.split()))
            if len(cen) != 2: raise ValueError("Exactly 2 coordinates were expected")

            rings = []
            closedrings = {}
            while True:
                dline = next(it)
                if dline == "$": break
                r = float(dline)
                print "Read r:=", r
                dline = next(it)
                closed = bool(int(dline))
                thetasegments = []
                while True:
                    dline = next(it)
                    if dline == "$": break
                    if dline != "new_segment:": raise ValueError("The command 'new_segment:' was expected")
                    thetas = []
                    while True:
                        dline = next(it)
                        if dline == "$": break
                        theta = float(dline)
                        thetas.append(theta)
                    thetasegments.append(thetas)
                rings.append((r, thetasegments))
                closedrings[r] = closed

            rays = []
            while True:
                dline = next(it)
                if dline == "$": break
                theta = float(dline)
                rs = []
                while True:
                    dline = next(it)
                    if dline == "$": break
                    r = float(dline)
                    rs.append(r)
                rays.append((theta, rs))

            enring = {}
            while True:
                dline = next(it)
                if dline == "$": break
                r, sd, d8, d10, e = map(float, dline.split())
                enring[r] = sd, d8, d10, e

            enray = {}
            while True:
                dline = next(it)
                if dline == "$": break
                theta = float(dline)
                enr = {}
                while True:
                    dline = next(it)
                    if dline == "$": break
                    r, sd, d8, d10, e = map(float, dline.split())
                    enr[r] = sd, d8, d10, e
                enray[theta] = enr

            self.energycache[cen] = rings, closedrings, rays, enring, enray


def addz(dtm, cen, rings, rays):
    "Add the z dimension to lines."
    crays = []
    for theta, rs in rays:
        c = cos(theta)
        s = sin(theta)
        radius = {}
        pline = []
        for r in rs:
            x = cen[0]+r*c
            y = cen[1]+r*s
            radius[x, y] = r
            pline.append((x, y, None))
        ni, cn = dtm.thanLineZ(pline)
        for i, (x, y, z) in enumerate(cn):
            cn[i] = x, y, z, radius.get((x, y))
        crays.append((theta, cn))

    crings = []
    for r, thetasegments in rings:
        cthetasegments = []
        for thetas in thetasegments:
            pline = [(cen[0]+r*cos(theta), cen[1]+r*sin(theta), None) for theta in thetas]
            ni, cn = dtm.thanLineZ(pline)
            cthetasegments.append(cn)
        crings.append((r, cthetasegments))
    return crings, crays


def penalties(dtm, rings, rays, crings, crays):
    "Calculate penalties for rings and rays; one penalty per ring, one penalty per ray."
    enring = {}                               #energy of ring road at radious r from the center
    for r, cthetasegments in crings:
        sd = d8 = d10 = e = 0.0
        for i, pline in enumerate(cthetasegments):
            sda, d8a, d10a, ea = roadut.road1energy(pline)
            sd += sda
            d8 += d8a
            d10 += d10a
            e += ea
        enring[r] = sd, d8, d10, e

    enray = {}                                #energy of ray road at angle th with respect to the center
    for th,pline in crays:
        en = roadut.road1energyDetail(pline)
        enr = {}
        for (x, y, z, r), en1 in zip(pline, en):
            if r is None: continue
            enr[r] = en1
        enray[th] = enr
#        print th
#        print "------------------------------------"
#        for r, en1 in enray[th].iteritems():
#            print r, en1
    return enring, enray

def write_cache_gen(fw, dr, dxy, dtheta, xymm, hull):
        "Write the general data of the cache to a file (like) object."
        fw.write("# dr dxy dtheta:\n%f   %f   %f\n" % (dr, dxy, dtheta))
        fw.write("# xymm:\n%f   %f   %f   %f\n" % tuple(xymm))
        fw.write("# hull:\n")
        for cc in hull:
            fw.write("%f   %f\n" % (cc[0], cc[1]))
        fw.write("$\n")


def write_cache_center1(fw, cen, rings, closedrings, rays, enring, enray):
            "Write the data of a single center of the cache to a file (like) object."
            fw.write("#==========================================================\n")
            fw.write("# Center:\n")
            fw.write("%f   %f\n" % cen)

            fw.write("# Rings defined by theta segments:\n")
            for r, thetasegments in rings:
                fw.write("# Radius:\n")
                fw.write("%f\n" % (r,))
                fw.write("# Closed:\n")
                fw.write("%d\n" % (closedrings[r],))
                for thetas in thetasegments:
                    fw.write("new_segment:\n")
                    fw.write("#   Thetas:\n")
                    for theta in thetas:
                        fw.write("    %f\n" % (theta,))
                    fw.write("$\n")
                fw.write("$\n")
            fw.write("$\n")

            fw.write("# Rays defined radiuses:\n")
            for theta, rs in rays:
                fw.write("# Theta:\n")
                fw.write("%f\n" % (theta,))
                fw.write("# Radiuses:\n")
                for r in rs:
                    fw.write("%f\n" % (r,))
                fw.write("$\n")
            fw.write("$\n")

            fw.write("# Rings: r   sd   d8   d10   e\n")
            rs = sorted(enring)
            for r in rs:
                fw.write("%f   %f   %f   %f   %f\n" % ((r,)+enring[r]))
            fw.write("$\n")

            fw.write("# Rays:\n")
            thetas = sorted(enray)
            for theta in thetas:
                fw.write("# theta:\n")
                fw.write("%f\n" % (theta,))
                fw.write("#     r   sd   d8   d10   e\n")
                enr = enray[theta]
                rs = sorted(enr)
                for r in rs:
                    fw.write("     %f   %f   %f   %f   %f\n" % ((r,)+enr[r]))
                fw.write("$\n")
            fw.write("$\n")


def test_center1(plot=False):
    "Test the geometry of a single center."
    import p_gvarcom, p_gdxf
    hull, cen, dr, dtheta, rays, rings, closedrings = geomcirc.test_center1(plot=False)
    dxf = p_gdxf.ThanDxfPlot()
    dxf.thanDxfPlots()
    geomcirc.plotcen(dxf, cen, hull, rays, rings)
    dxf.thanDxfPlot(0,0,999)
    dtm = p_gtri.ThanDTMlines()
    for cline in p_gvarcom.reSyk1(open("kam5k.syk")):
        dtm.thanAddLine1(cline)
    dtm.thanRecreate()
    crings, crays = addz(dtm, cen, rings, rays)
    enring, enray = penalties(dtm, rings, rays, crings, crays)

    cache = CircularCache(dr, dxy=10.0, factex=1.0, devdebug=True)
    cache.dtheta = dtheta
    cache.energycache = {}
    cache.energycache[cen] = rings, closedrings, rays, enring, enray
    cache.hull = hull
    testComputeEnergy(cache, cen)


def testComputeEnergy(cache, cen):
#    self.setPar(60.0, 140.0, 25.0+25.0, 30.0+30.0, 2.5+7.0+2.5, 2.5+9.0+2.5, True)
#    def setPar(self, boik1, boik2, hoik1, hoik2, bod1, bod2, biochange):
    s = p_ggen.Struct()
    boik1 = 60.0
    hoik1 = 30.0
    bod1 = 1.0
    s.boik = [boik1] * 20
    s.hoik = [hoik1] * 20
    s.bod  = [bod1]  * 20
    s.boik1 = boik1
    s.bod1 = bod1
    s.cen = cen
    d, d8, d10, e = cache.roadenergy(s)
    print d, d8, d10, e


def test_cache():
    "Test the cache with a given hull and dtm."
    import p_gvarcom
    dtm = p_gtri.ThanDTMlines()
    for cline in p_gvarcom.reSyk1(open("kam5k.syk")):
        dtm.thanAddLine1(cline)
    dtm.thanRecreate()
    hull = [(486161.360000, 4235170.651000),
            (485450.174000, 4235106.156000),
            (485436.229000, 4235360.649000),
            (485702.924000, 4235444.318000),
            (486098.608000, 4235489.639000),
            (486161.360000, 4235170.651000),
           ]
#    cache = CircularCache(dr=200.0, dxy=500.0, factex=1.0, devdebug=False)
    cache = CircularCache(dr=10.0, dxy=50.0, factex=1.0, devdebug=True)
    cache.build_cache(dtm, hull, prt=p_ggen.prg)
    print "Saving cache.."
    fw = open("kam5k.cache", "w")
    cache.write_cache(fw)
    fw.close()


def test_read_cache():
    "Test reading previousle saved cache."
    import p_gfil
    fr = open("kam5k.cache")
    fr = p_gfil.Datlin(fr)
    cache = CircularCache()
    cache.read_cache(fr)
    fw= open("q1.cache", "w")
    cache.write_cache(fw)
    fw.close()


if __name__ == "__main__":
#    test_center1()
#    test_cache()
    test_read_cache()
