# -*- coding: iso-8859-7 -*-
##############################################################################
# ThanCad 0.2.2 "Urban SAR": 2dimensional CAD with raster support for engineers.
# 
# Copyright (c) 2001-2013 Thanasis Stamos,  January 16, 2013
# URL:     http://thancad.sourceforge.net
# e-mail:  cyberthanasis@excite.com
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (www.gnu.org/licenses/gpl.html).
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##############################################################################

"""\
ThanCad 0.2.2 "Urban SAR": 2dimensional CAD with raster support for engineers.

Package which creates a bioclimatic city plan.
This module creates circluar city layouts.
"""
from math import cos, sin, hypot, ceil, tan, fabs
from p_ggen import xfrange, iterby2
from p_gmath import PI2, thanLineSeguw
import roadut


"""Module to find points of circle inside a region and determine
the intersection of the circle with the boundary of the region
if they exist.
"""

def radials(cc, hull, dtheta, dr):
    "The hull is convex polygon with the precondition hull[0] == hull[-1]."
    dis = hypot(cc[1], cc[0])      #Distance of center from the start
    rays = []
    for theta in xfrange(0.0, PI2, dtheta):
        theta = round(theta, 8)                             #Numerical reasons
        cb = cc[0]+dis*cos(theta), cc[1]+dis*sin(theta)     #Another point on the line (dis makes sure theat we do not lose precision in thanLineSeguw)
        rs = []
        for c1, c2 in iterby2(hull):
            ct = thanLineSeguw(cc, cb, c1, c2)
            if ct is None: continue
            u, _ = ct
#            print u*dis
            rs.append(u*dis)
        if len(rs) < 2: continue             #Theta is outside hull
        r = ceil(min(rs)/dr)*dr
        if r < 0.0: r = dr                   #The center is in the hull
        rmax = max(rs)
        rs = []
        while r < rmax:
            rs.append(r)
            r += dr
        if len(rs) > 1: rays.append((theta, rs))
    for i in xrange(len(rays)):
        dth = rays[i][0] - rays[i-1][0]
        if dth > 1.5*dtheta:                 #gap found
            if i == 0: break                 #Nothing to do; gap is in the right position
            rays = rays[i:] + rays[:i]
    else:
        pass                                 #No gap found; circular-> start at theta zero
    for i in xrange(1, len(rays)):
        dth = rays[i][0] - rays[i-1][0]
        assert dth <= 1.5*dtheta, "There should be only one gap in thetas for convex regions (found more)!"
    return rays


def arcsx(rays, dtheta, dr):
    "Find the arcs given the rays."
    _, rs = rays[0]
    rmin = rs[0]
    rmax = rs[-1]
    for _, rs in rays:
        rmin = min(rmin, rs[0])
        rmax = max(rmax, rs[-1])
    rmax + 0.1*dr                                #Numerical reasons
    rings = []
    closedrings = {}
    for r in xfrange(rmin, rmax, dr):
        thetasegments = []
        thetas = []
        for theta, rs in rays:
            if r in rs:
                thetas.append(theta)
            else:
                if len(thetas) > 1: thetasegments.append(thetas)
                thetas = []
        if len(thetas) > 1: thetasegments.append(thetas)
        if len(thetasegments) > 0:
            rings.append((r, thetasegments))
            c = False
            if len(thetasegments) == 1:
                thetas = thetasegments[0]
#                print r, "circular gap=", thetas[0]-thetas[-1], dpt(thetas[0]-thetas[-1])
                c = fabs(tan(thetas[0]-thetas[-1])) < tan(1.5*dtheta)
            closedrings[r] = c

    for i in xrange(1, len(rings)):
        dr2 = rings[i][0] - rings[i-1][0]
        assert dr2 <= 1.5*dr, "There should be no gap in the rings"
    return rings, closedrings


def plotcen(dxf, cen, hull, rays, rings):
    "Plot the circular city plan for a center."
    dxf.thanDxfSetLayer("cityplanregion")
    roadut.plotline(dxf, hull)

    dxf.thanDxfSetLayer("rays")
    dxf.thanDxfPlotPoint(cen[0], cen[1])
    for theta, rs in rays:
        c = cos(theta)
        s = sin(theta)
        pline = [(cen[0]+r*c, cen[1]+r*s) for r in rs]
        roadut.plotline(dxf, pline)

    dxf.thanDxfSetLayer("rings")
    for r, thetasegments in rings:
        for thetas in thetasegments:
            pline = [(cen[0]+r*cos(theta), cen[1]+r*sin(theta)) for theta in thetas]
            roadut.plotline(dxf, pline)


def test_center1(plot=True):
    "Test the geometry of a single center."
    hull = [(486161.360000, 4235170.651000),
            (485450.174000, 4235106.156000),
            (485436.229000, 4235360.649000),
            (485702.924000, 4235444.318000),
            (486098.608000, 4235489.639000),
            (486161.360000, 4235170.651000),
           ]
    cen=(485784.0,4235289.0)
    dtheta = PI2/100
    dr=10.0
    rays = radials(cen, hull, dtheta, dr)
    rings, closedrings = arcsx(rays, dtheta, dr)
    if plot:
        import p_gdxf
        dxf = p_gdxf.ThanDxfPlot()
        dxf.thanDxfPlots()
        plotcen(dxf, cen, hull, rays, rings)
        dxf.thanDxfPlot(0,0,999)
    return hull, cen, dr, dtheta, rays, rings, closedrings


if __name__ == "__main__":
    test_center1()
