#!/usr/bin/python
# -*- coding: iso-8859-7 -*-
import random, copy, itertools
import Image, ImageDraw
from math import cos, sin, pi, hypot, fabs
from p_ggen import prg, iterby2, xfrange, xfrangec, Struct
from p_gmath import thanLineSeg3, thanNear2
from p_gtri import ThanDTMlines, hull
from p_gchart import ThanChart, vis
from p_ganneal import SAAnnealable
import p_ggeom


def itercom(fr):
        "Iterates through a file ignoring comments."
        for dl in fr:
            dl = dl.strip()
            if dl == "" or dl[0] == "#": continue
            yield dl


class RoadGrid:
    "An object which contains a grid of road with nonuniform grid spaces."

    def __init__(self, hull, dtm):
        "Initialize a road grid."
        self.hull = hull
        self.dtm = dtm
        self.contours = None
        self.roads = ()
        self.roadenx = self.roadeny = None
        self.pxmin = self.pxmax = self.pymin = self.pymax = None
        self.uxmin = self.uxmax = self.uymin = self.uymax = None
        self.dx = self.dy = 1.0
        self.dtheta = 1.0
        self.changed = 0

    def __deepcopy__(self, memo):
        "This is to make shallow copy when ThanCad's dialog deepcopies values of widgets."
        return self.shallowClone()

    def shallowClone(self):
        "Make a shallow copy of self."
        r = RoadGrid(self.hull, self.dtm)
        r.contours = self.contours
        r.roads    = self.roads
        r.roadenx  = self.roadenx
        r.roadeny  = self.roadeny
        r.pxmin    = self.pxmin
        r.pxmax    = self.pxmax
        r.pymin    = self.pymin
        r.pymax    = self.pymax
        r.uxmin    = self.uxmin
        r.uxmax    = self.uxmax
        r.uymin    = self.uymin
        r.uymax    = self.uymax
        r.dx       = self.dx
        r.dy       = self.dy
        r.dtheta   = self.dtheta
        r.changed  = self.changed
        return r


    def __eq__(self, other):
        "Two objects are equal if the cache is the same."
        if not isinstance(other, self.__class__): return False
        return self.changed == other.changed


    def build_cache(self, prt=prg):
        "Create road grid every dx, dy, dtheta; cache the results."
#        self.dx = self.dy = 5.0       #Make the routine faster: for debugging
#        self.dtheta = 30.0            #Make the routine faster: for debugging
        self.roadenx = {}      #road energy in direction x (angle theta)
        self.roadeny = {}      #road energy in direction y (angle theta-90deg)
        self.pxmin = {}        #xmin of a road where x is the direction of angle theta and also the direction of the road
        self.pxmax = {}        #xmax 
        self.pymin = {}        #ymin of a road where y is the direction of angle theta-90deg and also the direction of the road
        self.pymax = {}        #ymax
        prt("Building cache..")
        for theta in xfrange(-90.0, 90.0, self.dtheta):
            prt("theta=%f" % theta)
            thrad = theta*pi/180.0
            pymin, pymax = self.cache_roads(self.roadenx, theta, thrad, self.dy)
            pxmin, pxmax = self.cache_roads(self.roadeny, theta, thrad-pi*0.5, self.dx)
            self.pymin[theta] = pymin
            self.pymax[theta] = pymax
            self.pxmin[theta] = pxmin
            self.pxmax[theta] = pxmax
        self.minmax(prt)
        self.changed += 1    #Object has been modified

    def minmax(self, prt=prg):
        "Find contours and xy minmax."
        prt("Determining contours inside city plan..")
        m = p_ggeom.MesaHull(self.hull)
        self.contours = [lin for lin in self.dtm.thanLines if m.mesa(lin[0]) or m.mesa(lin[1])]
        self.uxmin = min(itertools.chain((c[0] for c in lin for lin in self.contours), (c[0] for c in self.hull)))
        self.uxmax = max(itertools.chain((c[0] for c in lin for lin in self.contours), (c[0] for c in self.hull)))
        self.uymin = min(itertools.chain((c[1] for c in lin for lin in self.contours), (c[1] for c in self.hull)))
        self.uymax = max(itertools.chain((c[1] for c in lin for lin in self.contours), (c[1] for c in self.hull)))

    def cache_roads(self, roads, theta, thrad, dy):
        "Create grid; direction theta with respect to x-axis, varied widths,lenths of O.T. varies lengths of roads."
        t = cos(thrad), sin(thrad)
        n = -t[1], t[0]
        py = [n[0]*c[0]+n[1]*c[1] for c in self.hull]
        pymin = int(min(py)/dy+1.1)*dy     # Avoid py exactly on boundary
        pymax = int(max(py)/dy-1.1)*dy     # Avoid py exactly on boundary

        for pyaxis in xfrangec(pymin, pymax, dy):
            c1, c2 = thanPolygonLine(self.hull, pyaxis, n)
            assert c1 != None, "There should be 2 intersections!"
            nc, cprof = self.dtm.thanLineZ((c1, c2))
            if nc == -1:
                roads[theta, pyaxis] = None      # No profile was found
            else:
                roads[theta, pyaxis] = self.cache_roadenergy(cprof)
        return pymin, pymax


    def write_cache(self, fw):
        "Write the cache to a file."
        fw.write("# dx dy dtheta:\n%f   %f   %f\n" % (self.dx, self.dy, self.dtheta))
        for nam, pc in ("pxmin", self.pxmin), ("pxmax", self.pxmax), ("pymin", self.pymin), ("pymax", self.pymax):
            fw.write("\n# %s:\n" % nam)
            for key in sorted(pc.iterkeys()):
                fw.write("%f   %f\n" % (key, pc[key]))
            fw.write("$\n")
        for nam, pc in ("road energy x", self.roadenx), ("road energy y", self.roadeny):
            fw.write("\n# %s:\n" % nam)
            for key in sorted(pc.iterkeys()):
                val = pc[key]
                fw.write("%f  %f   %f  %f  %f  %f\n" % (key[0], key[1], val[0], val[1], val[2], val[3]))
            fw.write("$\n")


    def read_cache(self, fr):
        "Write the cache to a file."
        self.dx = self.dy = 1.0
        self.dtheta = 1.0
        self.roadenx = {}
        self.roadeny = {}
        self.pxmin = {}
        self.pxmax = {}
        self.pymin = {}
        self.pymax = {}
        it = itercom(fr)

        self.dx, self.dy, self.dtheta = map(float, it.next().split())

        for nam, pc in ("pxmin", self.pxmin), ("pxmax", self.pxmax), ("pymin", self.pymin), ("pymax", self.pymax):
            while True:
                dline = it.next()
                if dline == "$": break
                key, val = map(float, dline.split())
                pc[key] = val

        for nam, pc in ("road energy x", self.roadenx), ("road energy y", self.roadeny):
            while True:
                dline = it.next()
                if dline == "$": break
                key1, key2, val1, val2, val3, val4 = map(float, dline.split())
                pc[key1, key2] = val1, val2, val3, val4
        self.minmax()
        self.changed += 1     #Object has been modified


    w1 = 10.0
    w2 = 100.0
    def cache_roadenergy(self, cc):
        "Calculates the energy of a road."
        sd = e = 0.0
        d8 = 0.0
        d10 = 0.0
        for ca, cb in iterby2(cc):
                d = hypot(cb[0]-ca[0], cb[1]-ca[1])
                sd += d
                dz = fabs(cb[2]-ca[2])
                slope = dz*100.0/d
                if slope <= 8.0:
                    e += 0.0
                elif slope < 10.0:
                    e += d*self.w1*(slope-8.0)
                    d8 += d
                else:
                    e += d*self.w1*(10.0-8.0) + d*self.w2*(slope-10.0)
                    d8 += d
                    d10 += d
        return sd, d8, d10, e


    def iterOT(self, state):
        "Iterate through all OT."
        t = cos(state.theta*pi/180.0), sin(state.theta*pi/180.0)
        n = -t[1], t[0]
        px = [t[0]*c[0]+t[1]*c[1] for c in self.hull]
        py = [n[0]*c[0]+n[1]*c[1] for c in self.hull]
        pxmin = min(px)
        pxmax = max(px)
        pymin = min(py)
        pymax = max(py)

#        print "bod=", state.bod
#        print "boik=", state.boik
#        print "hoik=", state.hoik

        pya = pymin
        jy = jjy = 0
        while True:
            pya += state.bod[jjy]
            pyb = pya+state.boik[jjy]
            if pyb > pymax: break
            ca1, ca2 = thanPolygonLine(self.hull, pya, n)
            assert ca1 != None, "There should be 2 intersections!"
            cb1, cb2 = thanPolygonLine(self.hull, pyb, n)
            assert cb1 != None, "There should be 2 intersections!"
            pxa1 = t[0]*ca1[0]+t[1]*ca1[1]
            pxa2 = t[0]*ca2[0]+t[1]*ca2[1]
            if pxa1 > pxa2: ca1, pxa1, ca2, pxa2 = ca2, pxa2, ca1, pxa1
            pxb1 = t[0]*cb1[0]+t[1]*cb1[1]
            pxb2 = t[0]*cb2[0]+t[1]*cb2[1]
            if pxb1 > pxb2: cb1, pxb1, cb2, pxb2 = cb2, pxb2, cb1, pxb1
            px1min = max((pxa1, pxb1))
            px2max = min((pxa2, pxb2))

            px1 = pxmin
            jx = jjx = 0
            while True:
                px1 += state.bod[jjx]
                px2 = px1 + state.hoik[jjx]
                if px2 > px2max: break
                if px1 >= px1min:
                    cc = []
                    for px, py in (px1, pya), (px2, pya), (px2, pyb), (px1, pyb), (px1, pya):
                        cc.append([px*t[0]+py*n[0], px*t[1]+py*n[1]])
                    yield cc
                px1 = px2
                jx += 1
                if jx < len(state.hoik): jjx = jx

            pya = pyb
            jy += 1
            if jy < len(state.boik): jjy = jy


    def energy(self, s):
        "Calculates the energy of the system or the difference to the ideal configuration."
        sd, sd8, sd10, se = self.roadenergy(self.roadenx, self.dy, s.theta, self.pymin, self.pymax, s.boik, s.bod)
        d, d8, d10, e     = self.roadenergy(self.roadeny, self.dx, s.theta, self.pxmin, self.pxmax, s.hoik, s.bod)
        sd += d
        sd8 += d8
        sd10 += d10
        se += e
        return se/sd


    def energy3(self, s):
        "Calculates the energy of the system or the difference to the ideal configuration."
        sd, sd8, sd10, se = self.roadenergy(self.roadenx, self.dy, s.theta, self.pymin, self.pymax, s.boik, s.bod)
        d, d8, d10, e     = self.roadenergy(self.roadeny, self.dx, s.theta, self.pxmin, self.pxmax, s.hoik, s.bod)
        sd += d
        sd8 += d8
        sd10 += d10
        se += e
        return se/sd, sd8, sd10


    def roadenergy(self, roaden, dy, theta, pymin1, pymax1, boik, bod):
        "Find the energy of roads along x or y axis."
        py = pymin1[theta]
        pymax = pymax1[theta]
        j = jj = 0
        sd = sd8 = sd10 = se = 0.0
        while True:
            pyaxis = py + bod[jj]*0.5
            pyaxis = int(pyaxis/dy+0.5)*dy
            if pyaxis > pymax: break
            d, d8, d10, e = roaden[theta, pyaxis]
            sd += d
            sd8 += d8
            sd10 += d10
            se += e
            py += bod[jj] + boik[jj]
            j += 1
            if j < len(boik): jj = j
        return sd, sd8, sd10, se


    def roadcoor(self, s):
        "Find road coordinates along x and y axis."
        thrad = s.theta*pi/180.0
        roads = []
        self.roadcoor1(roads, self.dy, s.theta, thrad,        self.pymin, self.pymax, s.boik, s.bod)
        self.roadcoor1(roads, self.dx, s.theta, thrad-pi*0.5, self.pxmin, self.pxmax, s.hoik, s.bod)
        return roads


    def roadcoor1(self, roads, dy, theta, thrad, pymin1, pymax1, boik, bod):
        "Find road coordinates along x or y axis."
        t = cos(thrad), sin(thrad)
        n = -t[1], t[0]
        py = pymin1[theta]
        pymax = pymax1[theta]
        j = jj = 0
        while True:
            pyaxis = py + bod[jj]*0.5
            pyaxis = int(pyaxis/dy+0.5)*dy
            if pyaxis > pymax: break
            c1, c2 = thanPolygonLine(self.hull, pyaxis, n)
            assert c1 != None, "There should be 2 intersections!"
            roads.append((c1, c2))
            py += bod[jj] + boik[jj]
            j += 1
            if j < len(boik): jj = j


    def tochart(self, ch, hulls=(), iso=(), ot=(), roads=(), colot="blue"):
        "Show the road grid."
        for lines, col, _ in self.itlines(hulls, iso, ot, roads, colot):
            if lines == None: continue
            for line1 in lines:
                xx = [c[0] for c in line1]
                yy = [c[1] for c in line1]
                if lines is roads: ch.curveAdd(xx, yy, style="dasheddot", size=(12, -6, 3, -6), color=col)
                else:              ch.curveAdd(xx, yy, color=col)

    def itlines(self, hulls=(), iso=(), ot=(), roads=(), colot="blue"):
        "Iterate through the lines to be plotted."
        return [(hulls, "green", "hull"),
                (iso,   "brown", "contours"),
                (ot,    colot,   "ot"),
                (roads, "red",   "roads"),
               ]

    def topil(self, imd, ct, hulls=(), iso=(), ot=(), roads=(), colot="blue"):
        "Show the road grid."
        g2li = ct.global2Locali
        for lines, col, _ in self.itlines(hulls, iso, ot, roads, colot):
            if lines == None: continue
            for line1 in lines:
                cc = [g2li(c[0], c[1]) for c in line1]
                imd.line(cc, fill=col)

    def todxf(self, dxf, title="", hulls=(), iso=(), ot=(), roads=(), colot="blue", xor=0.0, yor=0.0):
        "Show the road grid."
        from p_gimdxf.thancolors import thanDxfColName2Rgb, thanRgb2DxfColCodeApprox
        for lines, col, laynam in self.itlines(hulls, iso, ot, roads, colot):
            if lines == None: continue
            for line1 in lines:
                xx = [xor+c[0] for c in line1]
                yy = [yor+c[1] for c in line1]
                _layer(dxf, laynam, col)
                dxf.thanDxfPlotPolyline(xx, yy)
        dy = (self.uymax-self.uymin)/100.0*5.0
        _layer(dxf, "TITLE", "white")
        dxf.thanDxfPlotSymbol(xor+self.uxmin+1.0*dy, yor+self.uymax+1.0*dy, dy, title, 0.0)

    def show(self, title="", iso=(), ot=(), roads=(), colot="blue"):
        "Show the road grid."
        ch = ThanChart(title)
        self.tochart(ch, (self.hull,), iso, ot, roads, colot)
        vis(ch, bg="white", title=title)

    def pilout(self, title="", width=600, height=400, iso=(), ot=(), roads=(), colot="blue"):
        "Plot the road grid to a dxf file."
        ct = self._pilct(width, height, (self.hull,), iso, ot, roads, colot)
        im = Image.new("RGB", (width, height), (255,255,255))
        imd = ImageDraw.Draw(im)
        self.topil(imd, ct, (self.hull,), iso, ot, roads)
        imd.text((5,1), text=title, fill=colot)
        return im, ct

    def _pilct(self, width=600, height=400, hulls=(), iso=(), ot=(), roads=(), colot="blue"):
        "Find the scale of the configuration to fill into an image of width x height."
        from p_gmath import ThanRectCoorTransf
        xmin = ymin = 1.0e100
        xmax = ymax = -ymin
        for lines, _, _ in self.itlines(hulls, iso, ot, roads, colot):
            if lines == None: continue
            for line1 in lines:
                for c in line1:
                    if c[0] < xmin: xmin = c[0]
                    if c[0] > xmax: xmax = c[0]
                    if c[1] < ymin: ymin = c[1]
                    if c[1] > ymax: ymax = c[1]
        assert xmin<xmax and ymin<ymax, "No hull, iso, ot, roads!"
        ct = ThanRectCoorTransf((xmin, ymin, xmax, ymax), (5, height-5, width-5, 5))
        return ct

    def dxfout(self, title="", iso=(), ot=(), roads=(), colot="blue"):
        "Plot the road grid to a dxf file."
        from p_gdxf import ThanDxfPlot
        dxf = ThanDxfPlot()
        dxf.thanDxfPlots()
        self.todxf(dxf, title, (self.hull,), iso, ot, roads, colot)
        dxf.thanDxfPlot(0,0,999)


def thanPolygonLine(cpol, pn, n):
    "Finds the intersection between convex polygon segment and line perpendicular to n, whose projection on n axis is pn."
    cts = []
    for i in xrange(len(cpol)):
        ct = thanLineSeg3(pn, n, cpol[i-1], cpol[i])
        if ct != None: cts.append(ct)
    if len(cts) == 0: return None, None
    cts.sort()
    j = 1
    while j < len(cts):
        i = j - 1
        if thanNear2(cts[j], cts[i]):
            del cts[j]
        else:
            j += 1
    assert len(cts) == 2, "There should be exactly 2 intersections"
    return cts[0], cts[1]


def lineprofile(clines):
    "Creates a (3D) line profile into a new drawing."
    dfact = 0.1
    charts = []
    for cp,pyaxis,col in clines:
        d = [0.0*dfact]
        for ca, cb in iterby2(cp):
            d.append(d[-1]+hypot(cb[1]-ca[1], cb[0]-ca[0])*dfact)
        zmin = min(ca[2] for ca in cp)
        ch = ThanChart()
        ch.curveAdd((d[0], d[-1]), (0.0, 0.0), color=col)
        ch.curveAdd((d[0], d[0]), (0.0, cp[0][2]-zmin))
        for i in xrange(1, len(cp)):
            ch.curveAdd((d[i], d[i], d[i-1]), (0.0, cp[i][2]-zmin, cp[i-1][2]-zmin), color=col)
        charts.append(ch)
    return charts


def _layer(dxf, lay, col):
    "Sets layer and color."
    from p_gimdxf.thancolors import thanDxfColName2Rgb, thanRgb2DxfColCodeApprox
    rgb = thanDxfColName2Rgb[col]
    icol = thanRgb2DxfColCodeApprox(rgb)
    dxf.thanDxfSetLayer(lay)
    dxf.thanDxfSetColor(icol)


class PolConfiguration(SAAnnealable):
    "A grid of roads in region amenable to simulated annealling."

    def __init__(self, hull=None, dtm=None):
        "Initial configuration is a single room."
        SAAnnealable.__init__(self)
        self.pol = RoadGrid(hull, dtm)
        self.state = Struct()
        self.setPar(50.0, 100.0, 8.0+15.0+8.0, 12.0+25.0+12.0, 2.0+10.0+2.0, 2.5+12.0+2.5, True)
        self.setPar(60.0, 140.0, 25.0+25.0, 30.0+30.0, 2.5+7.0+2.5, 2.5+9.0+2.5, True)
        self.ndim = 1+1


    def setPar(self, boik1, boik2, hoik1, hoik2, bod1, bod2, biochange):
        "Sets the parameters of city blocks."
        self.boik1 = boik1        # Parallel to ideal direction (east-west)
        self.boik2 = boik2        # Parallel to ideal direction (east-west)
        self.hoik1 = hoik1        # Perpendicular to ideal direction (tolerance+house_width1+tolerance+house_with2+tolerance))
        self.hoik2 = hoik2        # Perpendicular to ideal direction (north-south)
        self.bod1 = bod1          # Road width (pavement+width+pavement)
        self.bod2 = bod2
        self.state.e = -1.0             # Invalid value for energy
        self.state.boik = [self.boik1]
        self.state.hoik = [self.hoik1]
        self.state.bod = [self.bod1]
        self.biochange = biochange      # If true, bioclimatic constraints are forced
        if self.biochange: self.state.theta = 90.0   #Theta is angle of ideal direction + 90deg (shows to the north)
        else:              self.state.theta = 0.0    #Theta is angle of ideal direction + 90deg *shows to the north)


    def repairState(self):
        "Initalizes state after cache has been created or read."
        nb = int( (self.pol.pymax[0]-self.pol.pymin[0]) / (self.bod1+self.boik1) + 0.5)
        nh = int( (self.pol.pxmax[0]-self.pol.pxmin[0]) / (self.bod1+self.hoik1) + 0.5)
        print "nb=", nb, "nh=", nh
        self.state.boik = [self.boik1]*nb
        self.state.hoik = [self.hoik1]*nh
        self.state.bod = [self.bod1]*max(nb, nh)
        for i in xrange(len(self.state.bod)):                          #Thanasis2010_05_21
            self.state.bod[i] = self.r.uniform(self.bod1, self.bod2)   #Thanasis2010_05_21
        if self.biochange: self.state.theta = 90.0
        else:              self.state.theta = 0.0
        self.ndim = 2*(nb + nh)


    def test(self):
        "Tests energy for various angles."
        for theta in xfrange(-90, 90, 10):
            self.state.theta = theta*pi/180.0
            self.pol.create_geometry(self.state)
            print "theta=%8.1f   energy=%15.6f" % (theta, self.pol.energy())


    def energyState(self):
        "Computes the energy of the configuration."
        return self.pol.energy(self.state)


    def changeState(self):
        "Changes the configuration a little."
        p = self.r.random()
        if p < 0.20:      # Change theta
            theta = self.state.theta
            if self.biochange:
                if theta < 0.0: theta = self.state.theta = theta+180.0
            while True:
                dth = 5.0
                theta = self.state.theta + self.r.uniform(-dth, dth)
                theta = int(theta/self.pol.dtheta)*self.pol.dtheta
                if self.biochange:
                    if 90.0-22.5 <= theta < 90.0+22.5: break
                else:
                    if -90.0 <= theta < 90.0: break
            if self.biochange:
                if theta > 89.0: theta = theta-180.0
            self.state.theta = theta
        elif p < 0.4 :    # Change width of OT
            i = self.r.randint(0, len(self.state.boik)-1)
            self.state.boik[i] = int(self.r.uniform(self.boik1, self.boik2)/self.pol.dy)*self.pol.dy
        else:             # Change height of OT
            i = self.r.randint(0, len(self.state.hoik)-1)
            self.state.hoik[i] = int(self.r.uniform(self.hoik1, self.hoik2)/self.pol.dx)*self.pol.dx


    def getDimensions(self):
        "Return the dimensionality of current configuration of the annealing object."
        return self.ndim  # Dimensionality of the configuration

    def getState(self):
        "Return an object which fully reflects the state of the annealing object."
        return copy.deepcopy(self.state)

    def setState(self, state):
        "Replace current state of the anneling object with the one in variable state."
        self.state = state


    def initState_old(self):
      "Initialize random changes and calibrate energy."
      self.efact = 1.0
      ndim = self.ndim
      tries = ndim*10
      for i in xrange(tries):        # Randomise a bit
          j = self.changeState()
      e1 = self.energyState()
      de = 0.0
      ndim = self.ndim
      tries = ndim*10
      for i in xrange(tries):
          j = self.changeState()
          e2 = self.energyState()
          de += abs(e2-e1)
          e1 = e2
      de /= tries
      self.efact = 100.0 / de             # Normalise delta energy to 100: efact*de = 100
      prg(' =%.3f -> %3.f' % (e1, e1*self.efact))
      prg("   =%.3f -> %3.f" % (de, de*self.efact))
      print "efact=", self.efact
#      self.plot("First approximation")
      return self.ndim


    def wrstate(self, fw):
        "Save the current state to a file."
        s = self.state
#        e, d8, d10 = self.pol.energy3(s)
        fw.write("# Energy e, length with slope > 8%, length with slope > 10%:\n")
        fw.write("%f  %f  %f\n" % (s.e, s.d8, s.d10))
        fw.write("\n# theta:\n")
        fw.write("%f\n" % s.theta)
        for nam, b in ("boik", s.boik), ("hoik", s.hoik), ("bod", s.bod):
            fw.write("\n# %s:\n" % nam)
            for a in b: fw.write("%f\n" % a)
            fw.write("$\n")


    def rdstate(self, fr):
        "Reads the current state from a file."
        s = self.state
        it = itercom(fr)
        s.e, s.d8, s.d10 = map(float, it.next().split())
        s.theta = float(it.next())
        for nam, b in ("boik", s.boik), ("hoik", s.hoik), ("bod", s.bod):
            del b[:]
            while True:
                dline = it.next()
                if dline == "$": break
                b.append(float(dline))


    def imageForegroundState(self, im, ct, colot, T, e):
        "Superimpose image foreground to the given the background image."
        ot = list(self.pol.iterOT(self.state))
        imd = ImageDraw.Draw(im)
        self.pol.topil(imd, ct, ot=ot, colot=colot)
        if T != None: imd.text((5,1), text="t=%.2f  e=%.1f" % (T, e), fill=colot)

    def imageBackgroundState(self, imsize):
        "Superimpose background to the given (blank)image and return object for coordinate transformation."
        ot = list(self.pol.iterOT(self.state))
        return self.pol.pilout("", imsize[0], imsize[1], iso=self.pol.contours, roads=self.pol.roads)


if __name__ == "__main__":
    pass
