# -*- coding: iso-8859-7 -*-

##############################################################################
# ThanCad 0.1.2 "Decade": 2dimensional CAD with raster support for engineers.
# 
# Copyright (c) 2001-2011 Thanasis Stamos,  August 20, 2011
# URL:     http://thancad.sourceforge.net
# e-mail:  cyberthanasis@excite.com
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (www.gnu.org/licenses/gpl.html).
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##############################################################################

"""\
ThanCad 0.1.2 "Decade": 2dimensional CAD with raster support for engineers.

This module displays a dialog for the user to define the necessary elements
and options for automated floor plan architectural design.
"""

import sys, copy, ConfigParser, Tkinter

import p_gtkuti, p_gtkwid, p_ggen
from thantrans import Tarch as T
from thandiaarch import ThanArchCom


class ThanFplan(ThanArchCom):
    "Dialog for the pen thickness which the elements of a layer are plotted with."

    def __init__(self, master, vals=None, cargo=None, *args, **kw):
        "Extract initial rectification parameters."
        self.thanValsInit = vals           # This is structure not a scalar
        self.thanProj = cargo
        self.thanGps = []
        self.thanRel = []
        kw.setdefault("title", T["Automated Floor Plan Design"])
        kw.setdefault("buttonlabels", (T["Execute"], T["Cancel"]))
        ThanArchCom.__init__(self, master, *args, **kw)


    def thanValsDef(self):
        "Build default values."
        v = p_ggen.Struct()
        self.thanValsDefArch(v)   # Common Architectural parameters
        v.entMinRooms = 5
        v.entMaxRooms = 6

        v.entMintolWidth = 2.5
        v.entMinWidth = 3.0
        v.entMaxWidth = 6.0
        v.entMaxtolWidth = 7.0
        v.entMintolHeight = 3.0
        v.entMinHeight = 3.0
        v.entMaxHeight = 7.0
        v.entMaxtolHeight = 8.0

        v.entPenaltmore = 1.0
        v.entPenalt1 = 10.0
        v.entPenalty = 1.0
        return v


    def thanValsRead(self, v):
        "Reads the values from a file with specific suffix."
        fn = self.thanProj[0].parent / (self.thanProj[0].namebase + ".fpl")
        if not fn.exists(): return
        c = ConfigParser.SafeConfigParser()
        c.read(fn)
        tv = {}
        for (key,tit,wid,vld) in self.thanWids: tv[key] = tit, vld

        self.thanValsReadArch(c, tv, v)   # Common ICP parameters

        sec = "FLOOR PLAN CONSTRAINTS"
        for key in "entMinRooms entMaxRooms".split():
            tit, vld = tv[key]
            try:
                test = c.get(sec, tit)
                test = vld.thanValidate(test)
                if test != None: setattr(v, key, test)
            except:
                pass

        sec = "ROOM CONSTRAINTS"
        for key in "entMintolWidth entMinWidth entMaxWidth entMaxtolWidth "\
                   "entMintolHeight entMinHeight entMaxHeight entMaxtolHeight".split():
            tit, vld = tv[key]
            try:
                test = c.get(sec, tit)
                test = vld.thanValidate(test)
                if test != None: setattr(v, key, test)
            except:
                pass

        sec = "PENALTIES"
        for key in "entPenaltmore", "entPenalt1", "entPenalty":
            tit, vld = tv[key]
            try:
                test = c.get(sec, tit)
                test = vld.thanValidate(test)
                if test != None: setattr(v, key, test)
            except:
                pass


    def thanValsWrite(self, v):
        "Write the values to a file with specific suffix."
        fn = self.thanProj[0].parent / (self.thanProj[0].namebase + ".fpl")
        c = ConfigParser.SafeConfigParser()
        if fn.exists(): c.read(fn)
        tv = {}
        for (key,tit,wid,vld) in self.thanWids: tv[key] = tit, vld

        self.thanValsWriteArch(c, tv, v)   # Common ICP parameters

        sec = "FLOOR PLAN CONSTRAINTS"
        if not c.has_section(sec): c.add_section(sec)
        for key in "entMinRooms entMaxRooms".split():
            tit, vld = tv[key]
            c.set(sec, tit, str(getattr(v, key)))

        sec = "ROOM CONSTRAINTS"
        if not c.has_section(sec): c.add_section(sec)
        for key in "entMintolWidth entMinWidth entMaxWidth entMaxtolWidth "\
                   "entMintolHeight entMinHeight entMaxHeight entMaxtolHeight".split():
            tit, vld = tv[key]
            c.set(sec, tit, str(getattr(v, key)))

        sec = "PENALTIES"
        if not c.has_section(sec): c.add_section(sec)
        for key in "entPenaltmore", "entPenalt1", "entPenalty":
            tit, vld = tv[key]
            c.set(sec, tit, str(getattr(v, key)))

        try: c.write(fn.open("w"))
        except: pass


    def body(self, win):
        self.thanWids = []
        self.colfra = "blue"
#        self.option_add("*font", _fo)
        self.fraLogo(win, 0, theme=T["Automated Floor Plan Design Algorithms"], year=2010)
        self.fraGeom(win, 1)
        self.fraConstraints(win, 2)
        self.fraPenalties(win, 3)
        win.columnconfigure(0, weight=1)
        for (key,tit,wid,vld) in self.thanWids:
            setattr(self, key, wid)

        if self.thanValsInit == None:
            self.thanValsInit = self.thanValsDef()
            self.thanValsRead(self.thanValsInit)
        self.thanValsSaved = copy.deepcopy(self.thanValsInit)
        self.thanSet(self.thanValsInit)


    def fraConstraints(self, win, ir):
        "Select the room dimensions constraints."
        fra = Tkinter.Frame(win, bd=3, relief=Tkinter.RIDGE)
        fra.grid(row=ir, column=0, pady=5, sticky="we")

        lab = Tkinter.Label(fra, fg=self.colfra, text="%d."%(ir,))
        lab.grid(row=0, column=0)
        lab = Tkinter.Label(fra, anchor="w", fg=self.colfra, text=T["ROOM CONSTRAINTS:"])
        lab.grid(row=0, column=1, sticky="w", columnspan=4)

        keys = "entMintolWidth",            "entMinWidth",               "entMaxWidth",                "entMaxtolWidth"
        tits = "Min room length - hard (m)", "Min room length - soft (m)", "Max room length - soft (m)",  "Max room length - hard (m)"
        ir = 1
        for key, tit in zip(keys, tits):
            val = p_gtkwid.ThanValFloat(0.10, 1000.0)
            lab = Tkinter.Label(fra, text=T[tit])
            lab.grid(row=ir, column=1, sticky="e")
            wid = p_gtkwid.ThanEntry(fra, width=5)
            wid.grid(row=ir, column=2, sticky="we")
            self.thanWids.append((key, tit, wid, val))
            ir += 1

        keys = "entMintolHeight",           "entMinHeight",              "entMaxHeight",               "entMaxtolHeight"
        tits = "Min room width - hard (m)", "Min room width - soft (m)", "Max room width - soft (m)",  "Max room width - hard (m)"
        ir = 1
        for key, tit in zip(keys, tits):
            val = p_gtkwid.ThanValFloat(0.10, 1000.0)
            lab = Tkinter.Label(fra, text=T[tit])
            lab.grid(row=ir, column=4, sticky="e")
            wid = p_gtkwid.ThanEntry(fra, width=5)
            wid.grid(row=ir, column=5, sticky="we")
            self.thanWids.append((key, tit, wid, val))
            ir += 1

        wid = Tkinter.Frame(fra)
        wid.grid(row=2, column=3, sticky="we")

        fra.columnconfigure(3, weight=1)


    def fraPenalties(self, win, ir):
        "Select the room and roomconfiguration penalties."
        fra = Tkinter.Frame(win, bd=3, relief=Tkinter.RIDGE)
        fra.grid(row=ir, column=0, pady=5, sticky="we")

        lab = Tkinter.Label(fra, fg=self.colfra, text="%d."%(ir,))
        lab.grid(row=0, column=0)
        lab = Tkinter.Label(fra, anchor="w", fg=self.colfra, text=T["PENALTY (ADVANCED):"])
        lab.grid(row=0, column=1, sticky="w", columnspan=4)

        keys = "entPenaltmore", "entPenalt1", "entPenalty"
        tits = ("Number of rooms less(more) than min(max) (penalty/room)",
                "Room dimension less(more) than hard min(max) (penalty/m)",
                "Room dimension between soft and hard min(max) (penalty/m)"
               )
        expls = "(absolute)", "(normalised)", "(normalised)"
        ir = 1
        for key, tit, expl in zip(keys, tits, expls):
            val = p_gtkwid.ThanValFloat(0.10, 1000.0)
            lab = Tkinter.Label(fra, text=T[tit])
            lab.grid(row=ir, column=1, sticky="e")
            wid = p_gtkwid.ThanEntry(fra, width=5)
            wid.grid(row=ir, column=2, sticky="we")
            lab = Tkinter.Label(fra, text=T[expl])
            lab.grid(row=ir, column=3, sticky="w")
            self.thanWids.append((key, tit, wid, val))
            ir += 1

        wid = Tkinter.Frame(fra)
        wid.grid(row=2, column=4, sticky="we")

        fra.columnconfigure(4, weight=1)



    def validate(self, strict=True):
        """Returns true if the value chosen by the user is valid.

        If strict == True, then if an error is found, an error message is displayed,
        self.result is unchanged, and False is returned to the caller.
        If strict == True, and no errors are found, self.result is updated with
        the new values. True is returned to the caller.
        If strict == False, then if an error is found, a default value is used
        instead of the wrong one, self.results is set with the new values,
        and False is returned to the caller.
        If strict == False, and no errors are found, then, self.results is set
        with the new values, and True is returned to the caller.
        """
        ret, vs = ThanArchCom.validate(self, strict)
        if not ret and strict: return ret

        if vs.entMaxRooms < vs.entMinRooms:
            ret = False
            if strict:
                p_gtkuti.thanGudModalMessage(self, T["Number of rooms: max < min !"], T["Error in data"])
                return ret

        self.result = vs
        return ret
